import abc
import asyncio

from collections.abc import Iterable
from typing import Any

from .typing import Id, Record, Sample, SampleValue


class BaseDriver(metaclass=abc.ABCMeta):
    async def init(self) -> None:
        """Perform any initialization necessary to use this driver."""

        pass

    async def cleanup(self) -> None:
        """Perform any cleanups necessary to decommission this driver."""

        pass

    @abc.abstractmethod
    async def query(
        self,
        collection: str,
        fields: list[str] | None,
        filt: dict[str, Any],
        sort: list[tuple[str, bool]],
        limit: int | None,
    ) -> Iterable[Record]:
        """Return records from `collection`.

        Optionally project the returned records by only returning given `fields`, if not `None`.

        Filter records according to `filt`. Besides filtering by exact field value, the following filter operators are
        supported: `gt`, `ge`, `lt`, `le` and `in` (list of values).

        Sort results according to `sort` argument which is a list of pairs of fields and "descending" flags, if not an
        empty list.

        Optionally limit the number of returned records to `limit`, if not None."""

        return []

    @abc.abstractmethod
    async def insert(self, collection: str, record: Record) -> Id:
        """Insert `record` into `collection`.

        Return the associated record ID."""

        return "1"

    @abc.abstractmethod
    async def update(self, collection: str, record_part: Record, filt: dict[str, Any]) -> int:
        """Update records from `collection` with fields given in `record_part`.

        Only records that match the `filt` filter will be updated. Besides filtering by exact field value, the following
        filter operators are supported: `gt`, `ge`, `lt`, `le` and `in` (list of values).

        Return the total number of records that were updated."""

        return 0

    @abc.abstractmethod
    async def replace(self, collection: str, id_: Id, record: Record) -> bool:
        """Replace record with `id` in `collection`.

        Return `True` if matched and replaced."""

        return False

    @abc.abstractmethod
    async def remove(self, collection: str, filt: dict[str, Any]) -> int:
        """Remove records from `collection`.

        Only records that match the `filt` filter will be removed. Besides filtering by exact field value, the following
        filter operators are supported: `gt`, `ge`, `lt`, `le` and `in` (list of values).

        Return the total number of records that were removed."""

        return 0

    async def get_samples_slice(
        self,
        collection: str,
        obj_id: Id,
        from_timestamp: int | None,
        to_timestamp: int | None,
        limit: int | None,
        sort_desc: bool,
    ) -> Iterable[Sample]:
        """Return the samples of `obj_id` from `collection`.

        Filter results by an interval of time, if `from_timestamp` and/or `to_timestamp` are not `None`.
        `from_timestamp` is inclusive, while `to_timestamp` is exclusive.

        Optionally limit results to `limit` number of records, if not `None`. Limiting is done always with respect to
        the order of samples (i.e. chronologically ascending or descending).

        Sort the results by timestamp according to the value of `sort_desc`."""

        filt: dict[str, Any] = {
            "oid": obj_id,
        }

        if from_timestamp is not None:
            filt.setdefault("ts", {})["ge"] = from_timestamp

        if to_timestamp is not None:
            filt.setdefault("ts", {})["lt"] = to_timestamp

        sort = [("ts", sort_desc)]

        results = await self.query(collection, fields=None, filt=filt, sort=sort, limit=limit)

        return ((r["ts"], r["val"]) for r in results)

    async def get_samples_by_timestamp(
        self,
        collection: str,
        obj_id: Id,
        timestamps: list[int],
    ) -> Iterable[SampleValue]:
        """For each timestamp in `timestamps`, return the sample of `obj_id` from `collection` that was saved right
        before the (or at the exact) timestamp."""

        object_filter: dict[str, Any] = {
            "oid": obj_id,
        }

        query_tasks = []
        for timestamp in timestamps:
            filt = dict(object_filter, ts={"le": timestamp})
            task = self.query(collection, fields=None, filt=filt, sort=[("ts", True)], limit=1)
            query_tasks.append(task)

        task_results = await asyncio.gather(*query_tasks)

        samples = []
        for i, task_result in enumerate(task_results):
            query_results = list(task_result)
            if query_results:
                sample = query_results[0]["val"]
                samples.append(sample)
            else:
                samples.append(None)

        return samples

    async def save_sample(self, collection: str, obj_id: Id, timestamp: int, value: SampleValue) -> None:
        """Save a sample of an object with `obj_id` at a given `timestamp` to a specified `collection`."""

        record = {"oid": obj_id, "val": value, "ts": timestamp}

        await self.insert(collection, record)

    async def remove_samples(
        self,
        collection: str,
        obj_ids: list[Id] | None,
        from_timestamp: int | None,
        to_timestamp: int | None,
    ) -> int:
        """Remove samples from `collection`.

        If `obj_ids` is not `None`, only remove samples of given object ids.

        If `from_timestamp` and/or `to_timestamp` are not `None`, only remove samples within specified interval of time.
        `from_timestamp` is inclusive, while `to_timestamp` is exclusive.

        Return the number of removed samples."""

        filt: dict[str, Any] = {}
        if obj_ids:
            filt["oid"] = {"in": obj_ids}

        if from_timestamp is not None:
            filt.setdefault("ts", {})["ge"] = from_timestamp

        if to_timestamp is not None:
            filt.setdefault("ts", {})["lt"] = to_timestamp

        return await self.remove(collection, filt)

    def is_samples_supported(self) -> bool:
        """Tell whether samples are supported by this driver or not."""

        return False

    async def ensure_index(self, collection: str, index: list[tuple[str, bool]] | None) -> None:
        """Create an index on `collection` if not already present. `index` is a list of pairs of field names and
        "descending" direction flags.

        If `index` is `None`, the collection is assumed to be a collection of samples where the index is considered to
        be ascending on the timestamp field."""

        pass



import hashlib
import logging
import re
import time

from collections.abc import Callable

import jwt

from qtoggleserver import system
from qtoggleserver.conf import settings
from qtoggleserver.core.device import attrs as core_device_attrs


JWT_ISS = "qToggle"
JWT_ALG = "HS256"

ORIGIN_DEVICE = "device"
ORIGIN_CONSUMER = "consumer"

EMPTY_PASSWORD_HASH = hashlib.sha256(b"").hexdigest()

_AUTH_TOKEN_RE = re.compile(r"^Bearer\s+([a-z0-9_.-]+)$", re.IGNORECASE)

logger = logging.getLogger(__name__)


class AuthError(Exception):
    pass


def make_auth_header(origin: str, username: str | None, password_hash: str) -> str:
    claims: dict[str, str | int] = {"iss": JWT_ISS, "ori": origin}

    if username:
        claims["usr"] = username

    if system.date.has_real_date_time():
        claims["iat"] = int(time.time())

    token = jwt.encode(claims, key=password_hash or "", algorithm=JWT_ALG)

    return f"Bearer {token}"


def parse_auth_header(auth: str, origin: str, password_hash_func: Callable, require_usr: bool = True) -> str:
    m = _AUTH_TOKEN_RE.match(auth)
    if not m:
        raise AuthError("Invalid authorization header")

    # Decode but don't validate token yet
    token = m.group(1)

    try:
        payload = jwt.decode(
            token, algorithms=[JWT_ALG], options={"verify_signature": False}, leeway=settings.core.max_client_time_skew
        )
    except jwt.exceptions.InvalidTokenError as e:
        raise AuthError(f"Invalid JWT: {e}") from e

    # Validate claims
    if payload.get("iss") != JWT_ISS:
        raise AuthError("Missing or invalid iss claim in JWT")

    if payload.get("ori") != origin:
        raise AuthError("Missing or invalid ori claim in JWT")

    iat = payload.get("iat")
    if (iat is not None) and system.date.has_real_date_time():
        delta = time.time() - iat
        if abs(delta) > settings.core.max_client_time_skew:
            raise AuthError("JWT too old or too much in the future")

    usr = payload.get("usr")
    if require_usr:
        if not usr or not isinstance(usr, str):
            raise AuthError("Missing or invalid usr claim in JWT")

    # Validate username & signature
    password_hash = password_hash_func(usr)
    if not password_hash:
        raise AuthError(f"Unknown usr in JWT: {usr}")

    # Decode again to verify signature
    try:
        jwt.decode(
            token, key=password_hash, algorithms=[JWT_ALG], verify=True, leeway=settings.core.max_client_time_skew
        )
    except jwt.exceptions.InvalidSignatureError as e:
        raise AuthError("Invalid JWT signature") from e
    except jwt.exceptions.InvalidTokenError as e:
        raise AuthError(f"Invalid JWT: {e}") from e

    return usr


def consumer_password_hash_func(usr: str) -> str | None:
    if usr == "admin":
        return core_device_attrs.admin_password_hash
    elif usr == "normal":
        return core_device_attrs.normal_password_hash
    elif usr == "viewonly":
        return core_device_attrs.viewonly_password_hash
    else:
        return None



from qtoggleserver.core.typing import GenericJSONDict
from qtoggleserver.core.typing import PortValue as CorePortValue


class ExpressionException(Exception):
    pass


class ExpressionParseError(ExpressionException):
    def to_json(self) -> GenericJSONDict:
        return {"reason": "parse-error"}


class UnknownFunction(ExpressionParseError):
    def __init__(self, name: str, pos: int) -> None:
        self.name: str = name
        self.pos: int = pos

        super().__init__(f'Unknown function "{name}"')

    def to_json(self) -> GenericJSONDict:
        return {"reason": "unknown-function", "token": self.name, "pos": self.pos}


class InvalidNumberOfArguments(ExpressionParseError):
    def __init__(self, name: str, pos: int) -> None:
        self.name: str = name
        self.pos: int = pos

        super().__init__(f'Invalid number of arguments for function "{name}"')

    def to_json(self) -> GenericJSONDict:
        return {"reason": "invalid-number-of-arguments", "token": self.name, "pos": self.pos}


class InvalidArgumentKind(ExpressionParseError):
    def __init__(self, name: str, pos: int, num: int) -> None:
        self.name: str = name
        self.pos: int = pos
        self.num: int = num

        super().__init__(f'Invalid argument {num} kind for "{name}" at position {pos}')

    def to_json(self) -> GenericJSONDict:
        return {"reason": "invalid-argument-kind", "token": self.name, "pos": self.pos, "num": self.num}


class UnbalancedParentheses(ExpressionParseError):
    def __init__(self, pos: int) -> None:
        self.pos: int = pos

        super().__init__("Unbalanced parentheses")

    def to_json(self) -> GenericJSONDict:
        return {"reason": "unbalanced-parentheses", "pos": self.pos}


class UnexpectedEnd(ExpressionParseError):
    def __init__(self) -> None:
        super().__init__("Expression is unterminated")

    def to_json(self) -> GenericJSONDict:
        return {"reason": "unexpected-end"}


class CircularDependency(ExpressionParseError):
    def __init__(self, port_id: str) -> None:
        super().__init__(f'Expression creates a dependency loop via port "{port_id}"')

    def to_json(self) -> GenericJSONDict:
        return {"reason": "circular-dependency"}


class ExternalDependency(ExpressionParseError):
    def __init__(self, port_id: str, pos: int) -> None:
        self.port_id = port_id
        self.pos = pos

        super().__init__(f'External dependency on port "{port_id}"')

    def to_json(self) -> GenericJSONDict:
        return {"reason": "external-dependency", "token": self.port_id, "pos": self.pos}


class UnexpectedCharacter(ExpressionParseError):
    def __init__(self, c: str, pos: int) -> None:
        self.c = c
        self.pos: int = pos

        super().__init__(f'Unexpected character "{c}" at position {pos}')

    def to_json(self) -> GenericJSONDict:
        return {"reason": "unexpected-character", "token": self.c, "pos": self.pos}


class EmptyExpression(ExpressionParseError):
    def __init__(self) -> None:
        super().__init__("Expression is empty")

    def to_json(self) -> GenericJSONDict:
        return {"reason": "empty-expression"}


class ExpressionEvalError(ExpressionException):
    pass


class ValueUnavailable(ExpressionEvalError):
    def __init__(self, msg: str | None = None) -> None:
        super().__init__(msg or "Value is unavailable")


class InvalidArgumentValue(ExpressionEvalError):
    def __init__(self, arg_no: int, value: CorePortValue) -> None:
        self.arg_no: int = arg_no
        self.value: CorePortValue = value

        super().__init__(f"Invalid argument {self.arg_no}: {self.value}")


class UnknownPortId(ExpressionEvalError):
    def __init__(self, port_id: str) -> None:
        self.port_id = port_id

        super().__init__(f'Unknown port "{port_id}"')


class DisabledPort(ExpressionEvalError):
    def __init__(self, port_id: str) -> None:
        self.port_id = port_id

        super().__init__(f'Port "{port_id}" is disabled')


class PortValueUnavailable(ValueUnavailable):
    def __init__(self, port_id: str) -> None:
        self.port_id = port_id

        super().__init__(f'Port "{port_id}" value is unavailable')


class ExpressionArithmeticError(ExpressionEvalError):
    def __init__(self) -> None:
        super().__init__("Expression arithmetic error")


class EvalSkipped(ExpressionEvalError):
    def __init__(self) -> None:
        super().__init__("Evaluation skipped")



from qtoggleserver import system
from qtoggleserver.core import api as core_api
from qtoggleserver.core import device as core_device
from qtoggleserver.core import main
from qtoggleserver.core.api import schema as core_api_schema
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.core.device import events as core_device_events
from qtoggleserver.core.typing import Attributes


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def get_device(request: core_api.APIRequest) -> Attributes:
    return await core_device_attrs.to_json()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def put_device(request: core_api.APIRequest, params: Attributes) -> None:
    core_api_schema.validate(params, core_device_attrs.get_schema(loose=True))

    # Password fields must explicitly be ignored, so we pop them from supplied data
    for f in ("admin", "normal", "viewonly"):
        params.pop(f"{f}_password", None)

    # Ignore the date attribute
    params.pop("date", None)

    # Reset device attributes
    await core_device.reset(preserve_attrs=["admin_password_hash", "normal_password_hash", "viewonly_password_hash"])
    await core_device.load()

    try:
        await core_device_attrs.set_attrs(params, ignore_extra=True)
    except core_device_attrs.DeviceAttributeError as e:
        raise core_api.APIError(400, e.error, attribute=e.attribute)
    except Exception as e:
        raise core_api.APIError(500, "unexpected-error", message=str(e)) from e

    await core_device.save()
    await core_device_events.trigger_update()


@core_api.api_call(core_api.ACCESS_LEVEL_ADMIN)
async def patch_device(request: core_api.APIRequest, params: Attributes) -> None:
    def unexpected_field_code(field: str) -> str:
        if field in core_device_attrs.get_attrdefs():
            return "attribute-not-modifiable"
        else:
            return "no-such-attribute"

    core_api_schema.validate(
        params,
        core_device_attrs.get_schema(),
        unexpected_field_code=unexpected_field_code,
        unexpected_field_name="attribute",
    )

    try:
        reboot_required = await core_device_attrs.set_attrs(params)
    except core_device_attrs.DeviceAttributeError as e:
        raise core_api.APIError(400, e.error, attribute=e.attribute)
    except Exception as e:
        raise core_api.APIError(500, "unexpected-error", message=str(e)) from e

    await core_device.save()
    await core_device_events.trigger_update()

    if reboot_required:
        main.loop.call_later(2, system.reboot)



POST_PERIPHERALS = {
    "type": "object",
    "properties": {
        "driver": {
            "type": "string",
        },
        "name": {"oneOf": [{"type": "string", "pattern": "^[a-zA-Z_][a-zA-Z0-9_.-]{0,63}$"}, {"type": "null"}]},
    },
    "additionalProperties": True,
    "required": [
        "driver",
    ],
}

PUT_PERIPHERALS = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "driver": {
                "type": "string",
            },
            "name": {"oneOf": [{"type": "string", "pattern": "^[a-zA-Z_][a-zA-Z0-9_.-]{0,63}$"}, {"type": "null"}]},
            "id": {"oneOf": [{"type": "string", "pattern": "^[a-zA-Z_][a-zA-Z0-9_.-]{0,63}$"}, {"type": "null"}]},
        },
        "additionalProperties": True,
        "required": [
            "driver",
        ],
    },
}



from qtoggleserver import persist, system
from qtoggleserver.core import history as core_history
from qtoggleserver.core import ports as core_ports

from .base import EvalContext, EvalResult
from .exceptions import EvalSkipped, ExpressionEvalError, PortValueUnavailable
from .functions import Function, function
from .ports import PortRef


@function("AVAILABLE")
class AvailableFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    async def _eval(self, context: EvalContext) -> EvalResult:
        try:
            await self.args[0].eval(context)
            return True
        except ExpressionEvalError:
            return False


@function("DEFAULT")
class DefaultFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        try:
            return await self.args[0].eval(context)
        except ExpressionEvalError:
            return await self.args[1].eval(context)


@function("RISING")
class RisingFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: float | None = None

    async def _eval(self, context: EvalContext) -> EvalResult:
        value = await self.args[0].eval(context)

        result = False
        if self._last_value is not None and value > self._last_value:
            result = True
        self._last_value = value

        return result


@function("FALLING")
class FallingFunction(Function):
    MIN_ARGS = MAX_ARGS = 1

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: float | None = None

    async def _eval(self, context: EvalContext) -> EvalResult:
        value = await self.args[0].eval(context)

        result = False
        if self._last_value is not None and value < self._last_value:
            result = True
        self._last_value = value

        return result


@function("ACC")
class AccFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: float | None = None

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, accumulator = await self.eval_args(context)
        result = accumulator

        if self._last_value is not None:
            result += value - self._last_value

        self._last_value = value

        return result


@function("ACCINC")
class AccIncFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_value: float | None = None

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, accumulator = await self.eval_args(context)
        result = accumulator

        if (self._last_value is not None) and (value > self._last_value):
            result += value - self._last_value

        self._last_value = value

        return result


@function("HYST")
class HystFunction(Function):
    MIN_ARGS = MAX_ARGS = 3

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_result: int = 0

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, threshold1, threshold2 = await self.eval_args(context)

        self._last_result = int(
            (self._last_result == 0 and value > threshold2) or (self._last_result != 0 and value >= threshold1)
        )

        return self._last_result


@function("ONOFFAUTO")
class OnOffAutoFunction(Function):
    MIN_ARGS = MAX_ARGS = 2

    async def _eval(self, context: EvalContext) -> EvalResult:
        value, auto = await self.eval_args(context)
        if value > 0:
            return True
        elif value < 0:
            return False
        else:
            return auto


@function("SEQUENCE")
class SequenceFunction(Function):
    MIN_ARGS = 2
    DEPS = {"asap"}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._last_time_ms: int = 0

    async def _eval(self, context: EvalContext) -> EvalResult:
        if self._last_time_ms == 0:
            self._last_time_ms = context.now_ms

        args = await self.eval_args(context)
        num_values = len(args) // 2
        values = []
        delays = []
        total_delay = 0
        for i in range(num_values):
            values.append(args[i * 2])
            delays.append(args[i * 2 + 1])
            total_delay += args[i * 2 + 1]

        if len(delays) < len(values):
            delays.append(0)

        delta = context.now_ms - self._last_time_ms
        delta = delta % total_delay  # work modulo total_delay, to create repeat effect
        delay_so_far = 0
        result = values[0]
        for i in range(num_values):
            delay_so_far += delays[i]
            if delay_so_far >= delta:
                result = values[i]
                break

        return result


@function("LUT")
class LUTFunction(Function):
    MIN_ARGS = 5

    async def _eval(self, context: EvalContext) -> EvalResult:
        args = await self.eval_args(context)
        length = (len(args) - 1) // 2
        x = args[0]
        points = [(args[2 * i + 1], args[2 * i + 2]) for i in range(length)]
        points.sort(key=lambda p: p[0])

        if x < points[0][0]:
            return points[0][1]

        for i in range(length - 1):
            p1 = points[i]
            p2 = points[i + 1]

            if x > p2[0]:
                continue

            if x - p1[0] < p2[0] - x:  # closer to p1 than to p2
                return p1[1]
            else:
                return p2[1]

        return points[length - 1][1]


@function("LUTLI")
class LUTLIFunction(Function):
    MIN_ARGS = 5

    async def _eval(self, context: EvalContext) -> EvalResult:
        args = await self.eval_args(context)
        length = (len(args) - 1) // 2
        x = args[0]
        points = [(args[2 * i + 1], args[2 * i + 2]) for i in range(length)]
        points.sort(key=lambda p: p[0])

        if x < points[0][0]:
            return points[0][1]

        for i in range(length - 1):
            p1 = points[i]
            p2 = points[i + 1]

            if x > p2[0]:
                continue

            if p1[0] == p2[0]:
                return p1[1]

            return p1[1] + (p2[1] - p1[1]) * (x - p1[0]) / (p2[0] - p1[0])

        return points[length - 1][1]


@function("HISTORY")
class HistoryFunction(Function):
    MIN_ARGS = MAX_ARGS = 3
    DEPS = {"second"}
    ARG_KINDS = [PortRef]
    ENABLED = core_history.is_enabled

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._cached_sample: persist.Sample | None = None
        self._cached_timestamp: int = 0
        self._cached_max_diff: float | None = None

    async def _eval(self, context: EvalContext) -> EvalResult:
        if not system.date.has_real_date_time():
            raise EvalSkipped()

        args = await self.eval_args(context)
        port_id, timestamp, max_diff = args
        port = core_ports.get(port_id)
        assert port

        # Transform everything to milliseconds
        timestamp *= 1000
        max_diff *= 1000

        # If arguments have changed from last cached values, invalidate the sample
        if self._cached_timestamp != timestamp or self._cached_max_diff != max_diff:
            self._cached_sample = None

        if self._cached_sample is not None:
            return self._cached_sample[1]

        if max_diff > 0:
            # Look through all values after given timestamp, but no newer than timestamp + max_diff, and consider the
            # oldest one
            from_timestamp = timestamp
            to_timestamp = timestamp + max_diff
            sort_desc = False
            consider_curr_value = from_timestamp <= context.now_ms < to_timestamp
        elif max_diff < 0:
            # Look through all values before given timestamp, but no older than timestamp - abs(max_diff), and consider
            # the newest one
            if context.now_ms <= timestamp:
                # Sample from the future requested, the best we've got is current value
                from_timestamp = to_timestamp = None
                consider_curr_value = True
            else:
                # +1 is needed to satisfy the inclusion/exclusion of the interval boundaries
                from_timestamp = timestamp + max_diff + 1  # max_diff is negative
                to_timestamp = timestamp + 1
                consider_curr_value = False

            sort_desc = True
        else:  # assuming max_dif == 0
            # Look through all values after given timestamp and consider the oldest one
            from_timestamp = timestamp
            to_timestamp = None
            sort_desc = False
            consider_curr_value = from_timestamp <= context.now_ms

        if from_timestamp is not None or to_timestamp is not None:
            samples = await core_history.get_samples_slice(
                port, from_timestamp, to_timestamp, limit=1, sort_desc=sort_desc
            )
            samples = list(samples)
        else:
            samples = []

        if samples:
            self._cached_sample = samples[0]
            self._cached_timestamp = timestamp
            self._cached_max_diff = max_diff
            value = self._cached_sample[1]
        elif consider_curr_value:
            value = port.get_last_read_value()
        else:
            value = None

        if value is None:
            raise PortValueUnavailable(port.get_id())

        return value



import logging
import subprocess


logger = logging.getLogger(__name__)


def run_get_cmd(
    get_cmd: str,
    cmd_name: str | None = None,
    log_values: bool = True,
    exc_class: type | None = None,
    required_fields: list[str] | None = None,
) -> dict[str, str]:
    exc_class = exc_class or Exception

    try:
        config_bytes = subprocess.check_output(get_cmd, stderr=subprocess.STDOUT, shell=True)
    except Exception as e:
        raise exc_class(f"{cmd_name or get_cmd} get command failed: {e}") from e

    config = config_bytes.strip().decode()
    config_lines = config.split("\n")
    config_dict = {}
    for line in config_lines:
        line = line.strip()
        if not line:
            continue

        parts = line.split("=", 1)
        if len(parts) == 1:
            parts.append("")

        key, value = parts
        key = key.lower()[3:]  # strip leading "QS_"
        if value.startswith('"'):
            value = value[1:]
        if value.endswith('"') and not value.endswith('\\"'):
            value = value[:-1]

        config_dict[key] = value

    if cmd_name:
        if log_values:
            values_str = ", ".join(f'{k} = "{v}"' for k, v in sorted(config_dict.items()))
            logger.debug("got %s: %s", cmd_name, values_str)
        else:
            logger.debug("got %s", cmd_name)

    for field in required_fields or []:
        if field not in config_dict:
            msg = f"missing {field} field"
            if cmd_name:
                msg = f"invalid {cmd_name}: {msg}"

            raise exc_class(msg)

    return config_dict


def run_set_cmd(
    set_cmd: str, cmd_name: str | None = None, log_values: bool = True, exc_class: type | None = None, **config
) -> None:
    env = {f"QS_{k.upper()}": v for k, v in config.items()}

    exc_class = exc_class or Exception

    try:
        subprocess.check_output(set_cmd, env=env, stderr=subprocess.STDOUT, shell=True)
    except Exception as e:
        raise exc_class(f"{cmd_name or set_cmd} set command failed: {e}") from e

    if cmd_name:
        if log_values:
            values_str = ", ".join(f'{k} = "{v}"' for k, v in sorted(config.items()))
            logger.debug("%s set to: %s", cmd_name, values_str)
        else:
            logger.debug("%s set", cmd_name)



import asyncio
import errno
import os

from collections.abc import Callable
from functools import wraps
from typing import TextIO

from qtoggleserver.core import ports
from qtoggleserver.utils import json as json_utils


def _retry_after_configure(func: Callable) -> Callable:
    @wraps(func)
    async def wrapper(self, *args, **kwargs):  # noqa: ANN001, ANN202
        try:
            return await func(self, *args, **kwargs)
        except OSError as e:
            # If for some reason the GPIO controller has been disconnected and reconnected in the meantime,
            # we get an `ENODEV` and we retry after attempting a reconfiguration.
            if e.errno == errno.ENODEV:
                self.error("GPIO device missing (controller disconnected?), retrying")
                await asyncio.sleep(1)
                self._configure()
                return await func(self, *args, **kwargs)
            raise

    return wrapper


class GPIO(ports.Port):
    TYPE = ports.TYPE_BOOLEAN

    ADDITIONAL_ATTRDEFS = {
        "output": {
            "display_name": "Is Output",
            "description": "Controls the port direction.",
            "type": "boolean",
            "modifiable": True,
        }
    }

    BASE_PATH = "/sys/class/gpio"

    def __init__(self, no: int, def_value: bool | None = None, def_output: bool | None = None) -> None:
        self._no: int = no
        self._def_value: bool | None = def_value
        self._def_output: bool | None = def_output

        self._val_file: TextIO | None = None
        self._dir_file: TextIO | None = None

        super().__init__(port_id=f"gpio{no}")

    async def handle_enable(self) -> None:
        self._configure()

        if self._def_output is not None:
            await self.attr_set_output(self._def_output)

    @_retry_after_configure
    async def read_value(self) -> bool:
        self._val_file.seek(0)
        return self._val_file.read(1) == "1"

    @_retry_after_configure
    async def write_value(self, value: bool | None) -> None:
        if value is None:
            return

        self._val_file.seek(0)
        if value:
            value = "1"
        else:
            value = "0"

        self.debug('writing %s to "%s"', json_utils.dumps(value), self._val_file.name)
        self._val_file.write(value)
        self._val_file.flush()

    async def attr_is_writable(self) -> bool:
        return await self._is_output()

    async def attr_set_output(self, output: bool) -> None:
        if not self._dir_file:
            return

        self._dir_file.seek(0)

        if output:
            text = "out"
        else:
            text = "in"

        self.debug('writing "%s" to "%s"', text, self._dir_file.name)
        self._dir_file.write(text)
        self._dir_file.flush()

        if output and self._def_value is not None:
            await self.write_value(self._def_value)

    async def attr_is_output(self) -> bool:
        return await self._is_output()

    @_retry_after_configure
    async def _is_output(self) -> bool:
        if not self._dir_file:
            return False

        self._dir_file.seek(0)
        return self._dir_file.read(3) == "out"

    def _configure(self) -> None:
        self.debug("configuring GPIO")
        try:
            (self._val_file, self._dir_file) = self._open_files()
        except Exception as e:
            self.error("configuration failed: %s", e)
            raise

    def _open_files(self) -> tuple[TextIO, TextIO]:
        path = os.path.join(self.BASE_PATH, f"gpio{self._no}")

        if not os.path.exists(path):
            self.debug("exporting GPIO")
            with open(os.path.join(self.BASE_PATH, "export"), "w") as f:
                f.write(str(self._no))

        return open(os.path.join(path, "value"), "r+"), open(os.path.join(path, "direction"), "r+")



import asyncio
import inspect
import logging
import re

from collections.abc import Callable
from typing import Any

from tornado.iostream import StreamClosedError
from tornado.web import HTTPError, RequestHandler

from qtoggleserver.core import api as core_api
from qtoggleserver.core import responses as core_responses
from qtoggleserver.core.api import auth as core_api_auth
from qtoggleserver.core.device import attrs as core_device_attrs
from qtoggleserver.utils import json as json_utils


SESSION_ID_RE = re.compile(r"[a-zA-Z0-9]{1,32}")

logger = logging.getLogger(__name__)


class NoSuchFunction(HTTPError):
    def __init__(self) -> None:
        super().__init__(404, "no-such-function")


class BaseHandler(RequestHandler):
    _UNDEFINED: dict = {}

    def __init__(self, *args, **kwargs) -> None:
        self._json: Any = self._UNDEFINED
        self._response_body: str = ""
        self._response_body_json: Any = None

        super().__init__(*args, **kwargs)

    def get_request_json(self) -> Any:
        if self._json is self._UNDEFINED:
            try:
                self._json = json_utils.loads(self.request.body)
            except ValueError as e:
                logger.error("could not decode json from request body: %s", e)

                raise core_api.APIError(400, "malformed-body") from e

        return self._json

    def finish(self, chunk: str | None = None) -> asyncio.Future:
        self._response_body = chunk

        return super().finish(chunk)

    def finish_json(self, data: Any) -> asyncio.Future:
        self._response_body_json = data

        data = json_utils.dumps(data)
        data += "\n"

        self.set_header("Content-Type", "application/json; charset=utf-8")
        return self.finish(data)

    def get_response_body(self) -> str:
        return self._response_body

    def get_response_body_json(self) -> Any:
        return self._response_body_json

    def get_response_headers(self) -> dict[str, str]:
        return dict(self._headers.get_all())

    def get(self, **kwargs) -> None:
        raise NoSuchFunction()

    head = post = delete = patch = put = options = get

    def _handle_request_exception(self, exception: Exception) -> None:
        try:
            if isinstance(exception, HTTPError):
                logger.error("%s %s: %s", self.request.method, self.request.uri, exception)
                self.set_status(exception.status_code)
                self.finish_json(
                    {"error": (exception.log_message or getattr(exception, "reason", None) or str(exception))}
                )
            else:
                logger.error(str(exception), exc_info=True)
                self.set_status(500)
                self.finish_json({"error": "internal server error"})
        except RuntimeError:
            pass  # nevermind

    def data_received(self, chunk: bytes) -> None:
        pass


class APIHandler(BaseHandler):
    AUTH_ENABLED = True

    def __init__(self, *args, **kwargs) -> None:
        self.access_level: int = core_api.ACCESS_LEVEL_NONE
        self.username: str | None = None

        super().__init__(*args, **kwargs)

    def prepare(self) -> None:
        # Disable cache
        self.set_header("Cache-Control", "no-cache, no-store, must-revalidate, max-age=0")

        if not self.AUTH_ENABLED:
            return

        # Parse auth header
        auth = self.request.headers.get("Authorization")
        if auth:
            try:
                usr = core_api_auth.parse_auth_header(
                    auth, core_api_auth.ORIGIN_CONSUMER, core_api_auth.consumer_password_hash_func
                )
            except core_api_auth.AuthError as e:
                logger.warning(str(e))
                return
        else:
            if core_device_attrs.admin_password_hash == core_device_attrs.EMPTY_PASSWORD_HASH:
                logger.debug("authenticating request as admin due to empty admin password")
                usr = "admin"
            else:
                logger.warning("missing authorization header")
                return

        self.access_level = core_api.ACCESS_LEVEL_MAPPING[usr]
        self.username = usr

        logger.debug(
            "granted access level %s (username=%s)", core_api.ACCESS_LEVEL_MAPPING[self.access_level], self.username
        )

        # Validate session id
        session_id = self.request.headers.get("Session-Id")
        if session_id:
            if not SESSION_ID_RE.match(session_id):
                raise core_api.APIError(400, "invalid-header", header="Session-Id")

    async def call_api_func(self, func: Callable, default_status: int = 200, **kwargs) -> None:
        try:
            if self.request.method in ("POST", "PATCH", "PUT"):
                if self.request.headers.get("Content-Type", "").startswith("application/json"):
                    kwargs["params"] = self.get_request_json()
                else:
                    raise core_api.APIError(400, "invalid-header", header="Content-Type")

            response = func(self, **kwargs)
            if inspect.isawaitable(response):
                response = await response

            self.set_status(default_status)
            if response is not None or default_status == 200:
                await self.finish_json(response)
            else:
                await self.finish()
        except Exception as e:
            await self._handle_api_call_exception(func, kwargs, e)

    async def _handle_api_call_exception(self, func: Callable, kwargs: dict, error: Exception) -> None:
        kwargs = dict(kwargs)
        params = kwargs.pop("params", None)
        args = json_utils.dumps(kwargs)
        body = params and json_utils.dumps(params) or "{}"

        if isinstance(error, core_responses.HTTPError):
            error = core_api.APIError.from_http_error(error)

        if isinstance(error, core_api.APIError):
            logger.error("api call %s failed: %s (args=%s, body=%s)", func.__name__, error, args, body)

            self.set_status(error.status)
            if not self._finished:  # avoid finishing an already finished request
                await self.finish_json(error.to_json())

        if isinstance(error, core_api.APIAccepted):
            self.set_status(202)
            if not self._finished and error.response is not None:  # avoid finishing an already finished request
                await self.finish_json(error.response)
        elif isinstance(error, StreamClosedError) and func.__name__ == "get_listen":
            logger.debug("api call get_listen could not complete: stream closed")
        else:
            logger.error("api call %s failed: %s (args=%s, body=%s)", func.__name__, error, args, body, exc_info=True)

            self.set_status(500)
            if not self._finished:  # avoid finishing an already finished request
                await self.finish_json({"error": str(error)})



import logging
import os
import shutil

from typing import Any

from qtoggleserver.conf import settings
from qtoggleserver.utils import conf as conf_utils


logger = logging.getLogger(__name__)


def can_write_conf_file() -> bool:
    if not settings.source:
        return False

    return os.access(settings.source, os.W_OK)


def conf_file_to_dict() -> dict[str, Any]:
    if not can_write_conf_file():
        raise Exception("Configuration file not available")

    assert settings.source

    return conf_utils.config_from_file(settings.source)


def conf_file_from_dict(d: dict[str, Any]) -> None:
    if not can_write_conf_file():
        raise Exception("Configuration file not available")

    logger.debug("updating configuration file %s", settings.source)

    assert settings.source

    existing_d = conf_utils.config_from_file(settings.source)
    existing_d.update(d)

    config = conf_utils.config_from_dict(existing_d)
    config_str = conf_utils.config_to_str(config)

    # Create a backup
    try:
        shutil.copy(settings.source, f"{settings.source}.bak")
    except Exception:
        logger.warning("failed to create backup file %s", f"{settings.source}.bak", exc_info=True)

    with open(settings.source, "w") as f:
        f.write(config_str)



PUT_SLAVE_DEVICES = {"type": "array", "items": {"type": "object"}}

POST_SLAVE_DEVICES = {
    "type": "object",
    "properties": {
        "scheme": {"enum": ["http", "https"]},
        "host": {"type": "string", "maxLength": 256},
        "port": {"type": "integer", "min": 0, "max": 65535},
        "path": {"type": "string", "maxLength": 256},
        "admin_password": {"type": "string", "maxLength": 32},
        "admin_password_hash": {"type": "string", "minLength": 64, "maxLength": 64},
        "poll_interval": {"type": "number", "min": 0, "max": 86400},
        "listen_enabled": {"type": "boolean"},
    },
    "additionalProperties": False,
    "required": ["scheme", "host", "port", "path"],
}

PATCH_SLAVE_DEVICE = {
    "type": "object",
    "properties": {
        "enabled": {"type": "boolean"},
        "poll_interval": {"type": "number", "min": 0, "max": 86400},
        "listen_enabled": {"type": "boolean"},
    },
    "additionalProperties": False,
}

POST_SLAVE_DEVICE_EVENTS = {
    "type": "object",
    "properties": {"type": {"type": "string"}, "params": {"type": "object"}},
    "additionalProperties": False,
    "required": ["type"],
}

PATCH_DISCOVERED_DEVICE = {
    "type": "object",
    "properties": {"attrs": {"type": "object"}},
    "additionalProperties": False,
    "required": ["attrs"],
}



from __future__ import annotations

from qtoggleserver.core import ports as core_ports

from .base import EvalContext, EvalResult, Expression
from .exceptions import CircularDependency
from .literalvalues import LiteralValue


__all__ = [
    "ROLE_FILTER",
    "ROLE_TRANSFORM_READ",
    "ROLE_TRANSFORM_WRITE",
    "ROLE_VALUE",
    "EvalContext",
    "EvalResult",
    "Expression",
    "Function",
    "check_loops",
    "parse",
]

ROLE_VALUE = 1
ROLE_TRANSFORM_READ = 2
ROLE_TRANSFORM_WRITE = 3
ROLE_FILTER = 4

# A time jump of more than one day will prevent the evaluation of expressions such as time-processing
TIME_JUMP_THRESHOLD = 86_400_000


def parse(self_port_id: str | None, sexpression: str, role: int, pos: int = 1) -> Expression:
    while sexpression and sexpression[0].isspace():
        sexpression = sexpression[1:]
        pos += 1

    while sexpression and sexpression[-1].isspace():
        sexpression = sexpression[:-1]

    if sexpression and sexpression[0] in ("$", "@"):
        return PortExpression.parse(self_port_id, sexpression, role, pos)
    elif "(" in sexpression or ")" in sexpression:
        return Function.parse(self_port_id, sexpression, role, pos)
    else:
        return LiteralValue.parse(self_port_id, sexpression, role, pos)


async def check_loops(port: core_ports.BasePort, expression: Expression) -> None:
    seen_ports = {port}

    async def check_loops_rec(level: int, e: Expression) -> int:
        if isinstance(e, PortValue):
            p = e.get_port()
            if not p:
                return 0

            # A loop is detected when we stumble upon the initial port at a level deeper than 1
            if port is p and level > 1:
                return level

            # Avoid visiting the same port twice
            if p in seen_ports:
                return 0

            seen_ports.add(p)

            expr = p.get_expression()
            if expr:
                lv = await check_loops_rec(level + 1, expr)
                if lv:
                    return lv

            return 0
        elif isinstance(e, Function):
            for arg in e.args:
                lv = await check_loops_rec(level, arg)
                if lv:
                    return lv

        return 0

    if await check_loops_rec(1, expression) > 1:
        raise CircularDependency(port.get_id())


from .functions import Function  # noqa: E402
from .ports import PortExpression, PortValue  # noqa: E402



from typing import Any

from qtoggleserver.core import api as core_api
from qtoggleserver.core import responses as core_responses


# We can't use proper type annotations for slaves in this module because that would create unsolvable circular imports.
# Therefore, we use "Any" type annotation for Slave instances.


class SlaveError(Exception):
    pass


class InvalidDevice(SlaveError):
    pass


class NoListenSupport(SlaveError):
    def __init__(self, name: str) -> None:
        self.name: str = name

        super().__init__(f"Device {name} has no listen support")


class DeviceAlreadyExists(SlaveError):
    def __init__(self, name: str) -> None:
        self.name: str = name

        super().__init__(f"Device {name} already exists")


class DeviceRenamed(SlaveError):
    def __init__(self, slave: Any) -> None:
        self.slave = slave

        super().__init__(f"{slave} renamed")


class DeviceOffline(SlaveError):
    def __init__(self, slave: Any) -> None:
        self.slave = slave

        super().__init__(f"{slave} is offline")


class PortNotFound(SlaveError):
    def __init__(self, slave: Any, id_: str) -> None:
        self.slave = slave
        self.id: str = id_

        super().__init__(f"Could not find port {id_} of {slave}")


def adapt_api_error(error: Exception) -> Exception:
    if isinstance(
        error, (core_responses.HostUnreachable, core_responses.NetworkUnreachable, core_responses.UnresolvableHostname)
    ):
        return core_api.APIError(502, "unreachable")
    elif isinstance(error, core_responses.ConnectionRefused):
        return core_api.APIError(502, "connection-refused")
    elif isinstance(error, core_responses.InvalidJson):
        return core_api.APIError(502, "invalid-device")
    elif isinstance(error, core_responses.Timeout):
        return core_api.APIError(504, "device-timeout")
    elif isinstance(error, core_responses.HTTPError):
        return core_api.APIError.from_http_error(error)
    elif isinstance(error, DeviceOffline):
        return core_api.APIError(503, "device-offline")
    elif isinstance(error, InvalidDevice):
        return core_api.APIError(502, "invalid-device")
    elif isinstance(error, core_responses.AuthError):
        return core_api.APIError(400, "forbidden")  # yes, 400, since it's a slave authorization issue
    elif isinstance(error, core_api.APIError):
        return error
    else:  # leave error unchanged since it's probably an internal exception
        return error



import asyncio
import logging

from typing import Any

from qtoggleserver import persist
from qtoggleserver.conf import settings
from qtoggleserver.utils import dynload as dynload_utils

from .exceptions import DuplicatePeripheral, NoSuchDriver
from .peripheral import Peripheral
from .peripheralport import PeripheralPort


__all__ = ["Peripheral", "PeripheralPort"]


logger = logging.getLogger(__name__)

_registered_peripherals: dict[str, Peripheral] = {}


def get_all() -> list[Peripheral]:
    return list(_registered_peripherals.values())


def get(peripheral_id: str) -> Peripheral | None:
    return _registered_peripherals.get(peripheral_id)


async def add(peripheral_params: dict[str, Any], static: bool = False) -> Peripheral:
    params = dict(peripheral_params)
    class_path = params.pop("driver")
    params.pop("static", None)

    logger.debug('creating peripheral with driver "%s"', class_path)
    try:
        peripheral_class = dynload_utils.load_attr(class_path)
    except Exception:
        raise NoSuchDriver(class_path)

    p = peripheral_class(params=peripheral_params, static=static, **params)
    if p.get_id() in _registered_peripherals:
        raise DuplicatePeripheral(f"Peripheral {p.get_id()} already exists")

    p.debug("initializing")
    await p.handle_init()
    _registered_peripherals[p.get_id()] = p

    if not static:
        peripheral_params = dict(peripheral_params)
        peripheral_params.pop("static", None)
        await persist.replace("peripherals", p.get_id(), peripheral_params)

    return p


async def remove(peripheral_id: str, persisted_data: bool = True) -> None:
    p = _registered_peripherals[peripheral_id]

    p.debug("cleaning up")
    await p.handle_cleanup()
    _registered_peripherals.pop(peripheral_id)

    logger.debug("peripheral %s removed", peripheral_id)

    if persisted_data:
        await persist.remove("peripherals", filt={"id": peripheral_id})


async def init() -> None:
    logger.debug("loading static peripherals")
    for peripheral_params in settings.peripherals:
        try:
            await add(peripheral_params, static=True)
        except Exception:
            logger.error("failed to load peripheral %s", peripheral_params.get("driver"), exc_info=True)

    logger.debug("loading dynamic peripherals")
    for peripheral_params in await persist.query("peripherals"):
        try:
            await add(peripheral_params)
        except Exception:
            logger.error("failed to load peripheral %s", peripheral_params.get("driver"), exc_info=True)


async def cleanup() -> None:
    tasks = [asyncio.create_task(remove(p_id, persisted_data=False)) for p_id in _registered_peripherals.keys()]
    if tasks:
        await asyncio.wait(tasks)



import abc
import asyncio
import logging

from typing import cast

from qtoggleserver.core import main
from qtoggleserver.core.typing import AttributeDefinition, AttributeDefinitions
from qtoggleserver.peripherals import Peripheral, PeripheralPort


READ_INTERVAL_ATTRDEF = {
    "display_name": "Read Interval",
    "description": "How often to read peripheral data (set to 0 to disable reading).",
    "type": "number",
    "modifiable": True,
    "integer": True,
}


class PolledPeripheral(Peripheral, metaclass=abc.ABCMeta):
    DEFAULT_POLL_INTERVAL = 1800
    DEFAULT_RETRY_COUNT = 0
    DEFAULT_RETRY_POLL_INTERVAL = 60

    logger = logging.getLogger(__name__)

    def __init__(self, *, retry_poll_interval: int = -1, retry_count: int = -1, **kwargs) -> None:
        self._polling: bool = False
        self._poll_stopped: bool = False
        self._poll_task: asyncio.Task | None = None
        self._poll_interval: int = self.DEFAULT_POLL_INTERVAL
        self._poll_error: Exception | None = None
        self._retry_poll_interval: int = (
            retry_poll_interval if retry_poll_interval >= 0 else self.DEFAULT_RETRY_POLL_INTERVAL
        )
        self._retry_count: int = retry_count if retry_count >= 0 else self.DEFAULT_RETRY_COUNT
        self._retry_counter: int = 0

        super().__init__(**kwargs)

    async def _poll_loop(self) -> None:
        self.debug("polling started")
        self._polling = True
        self._retry_counter = 0

        while self._polling:
            try:
                if self._poll_interval == 0:
                    await asyncio.sleep(1)
                    continue
                if not main.loop:
                    # Skip polling during startup, while `main.loop` isn't initialized yet
                    await asyncio.sleep(1)
                    continue

                try:
                    await self.poll()
                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    retry_poll_interval = min(self._retry_poll_interval, self._poll_interval)
                    self.error(
                        "polling failed (retrying in %s seconds, %s/%s): %s",
                        retry_poll_interval,
                        self._retry_counter + 1,
                        self._retry_count,
                        e,
                        exc_info=True,
                    )
                    self._poll_error = e
                    self._retry_counter += 1
                    if self._retry_counter > self._retry_count:
                        self.set_online(False)
                    await asyncio.sleep(retry_poll_interval)
                    continue

                # Clear poll error, as the poll call has been successful
                self._poll_error = None
                self._retry_counter = 0
                self.set_online(True)

                # Granular sleep so it can be adjusted
                orig_poll_interval = self._poll_interval
                for i in range(self._poll_interval):
                    if not self._polling:
                        break

                    if orig_poll_interval != self._poll_interval:  # poll interval changed
                        break

                    await asyncio.sleep(1)
            except asyncio.CancelledError:
                self.debug("polling task cancelled")
                break

        self._poll_task = None
        self.debug("polling stopped")

    def set_poll_interval(self, interval: int) -> None:
        self._poll_interval = interval
        self.trigger_port_update_fire_and_forget(save=True)

    def get_poll_interval(self) -> int:
        return self._poll_interval

    def get_poll_error(self) -> Exception:
        return self._poll_error

    def has_poll_error(self) -> bool:
        return bool(self._poll_error)

    @abc.abstractmethod
    async def poll(self) -> None:
        raise NotImplementedError()

    async def handle_enable(self) -> None:
        self._poll_task = asyncio.create_task(self._poll_loop())

    async def handle_disable(self) -> None:
        if self._poll_task:
            self._polling = False  # will stop poll loop
            self._poll_task.cancel()
            await self._poll_task

        self._poll_error = None
        self._retry_counter = 0

    async def handle_cleanup(self) -> None:
        await super().handle_cleanup()

        if self._poll_task and not self._poll_task.done():
            self._polling = False
            self._poll_task.cancel()
            await self._poll_task


class PolledPort(PeripheralPort, metaclass=abc.ABCMeta):
    # Set these to None to disable read interval attribute
    READ_INTERVAL_MIN = 0
    READ_INTERVAL_MAX = 86400
    READ_INTERVAL_STEP = 1
    READ_INTERVAL_MULTIPLIER = 1
    READ_INTERVAL_UNIT = None

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        # Add read interval attrdef

    async def get_additional_attrdefs(self) -> AttributeDefinitions:
        attrdefs: AttributeDefinitions = {}
        if self.READ_INTERVAL_MIN is not None:
            attrdef: AttributeDefinition = dict(READ_INTERVAL_ATTRDEF)

            unit = self.READ_INTERVAL_UNIT
            if unit is None:
                if self.READ_INTERVAL_MULTIPLIER == 3600:
                    unit = "hours"
                elif self.READ_INTERVAL_MULTIPLIER == 60:
                    unit = "minutes"
                else:
                    unit = "seconds"

            attrdef.update(
                unit=unit,
                step=self.READ_INTERVAL_STEP,
                min=self.READ_INTERVAL_MIN,
                max=self.READ_INTERVAL_MAX,
            )

            attrdefs["read_interval"] = attrdef

        return attrdefs

    async def attr_set_read_interval(self, interval: int) -> None:
        peripheral = cast(PolledPeripheral, self.get_peripheral())
        peripheral.set_poll_interval(int(interval) * self.READ_INTERVAL_MULTIPLIER)

    async def attr_get_read_interval(self) -> int:
        peripheral = cast(PolledPeripheral, self.get_peripheral())
        return peripheral.get_poll_interval() // self.READ_INTERVAL_MULTIPLIER
